// AForge Image Processing Library
// AForge.NET framework
//
// Copyright  Andrew Kirillov, 2005-2008
// andrew.kirillov@gmail.com
//

namespace AForge.Imaging
{
    using System;
    using System.Drawing;
    using BrainTechLLC;
    using BrainTechLLC.ThreadSafeObjects;
    using System.Collections.Generic;
    using System.Threading;
    using CPPClassLibrary;
    using System.Windows.Forms;

    /// <summary>
    /// RGB components.
    /// </summary>
    /// 
    /// <remarks><para>The class encapsulates <b>RGB</b> color components.</para>
    /// <para><note><see cref="System.Drawing.Imaging.PixelFormat">PixelFormat.Format24bppRgb</see>
    /// actually means BGR format.</note></para>
    /// </remarks>
    /// 
    [Serializable]
    public class RGB
    {
        public static ImageFiltering filtering = new ImageFiltering();

        /// <summary>
        /// Index of red component.
        /// </summary>
        public const short R = 2;

        /// <summary>
        /// Index of green component.
        /// </summary>
        public const short G = 1;

        /// <summary>
        /// Index of blue component.
        /// </summary>
        public const short B = 0;

        /// <summary>
        /// Index of alpha component for ARGB images.
        /// </summary>
        public const short A = 3;

        /// <summary>
        /// Red component.
        /// </summary>
        public byte Red;

        /// <summary>
        /// Green component.
        /// </summary>
        public byte Green;

        /// <summary>
        /// Blue component.
        /// </summary>
        public byte Blue;

        public unsafe void ToHSL(HSL hsl)
        {
            int hue;
            double sat;
            double lum;
            ImageFiltering.DoRGBToHSL(Red, Green, Blue, &hue, &sat, &lum);

            hsl.Hue = hue;
            hsl.Saturation = sat;
            hsl.Luminance = lum;
            //int key = ((int)this.Red << 16) + ((int)this.Green << 8) + (int)this.Blue;
            //HSL temp;

            //// convert to HSL
            //// AForge.Imaging.HSL.FromRGB( rgb, hsl );
            //if (!AForge.Imaging.HSL.RGB_HSL.TryGetValue(key, out temp))
            //{
            //    temp = AForge.Imaging.HSL.CalcAndAddHSLFromRGB(this, key, true);
            //}

            //if (temp == null)
            //{
            //    temp = AForge.Imaging.HSL.CalcAndAddHSLFromRGB(this, key, false);
            //}

            //hsl.Hue = temp.Hue;
            //hsl.Luminance = temp.Luminance;
            //hsl.Saturation = temp.Saturation;
        }

        /// <summary>
        /// <see cref="System.Drawing.Color">Color</see> value of the class.
        /// </summary>
        public System.Drawing.Color Color
        {
            get { return Color.FromArgb(Red, Green, Blue); }
            set
            {
                Red = value.R;
                Green = value.G;
                Blue = value.B;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="RGB"/> class.
        /// </summary>
        public RGB() { }

        /// <summary>
        /// Initializes a new instance of the <see cref="RGB"/> class.
        /// </summary>
        /// 
        /// <param name="red">Red component.</param>
        /// <param name="green">Green component.</param>
        /// <param name="blue">Blue component.</param>
        /// 
        public RGB(byte red, byte green, byte blue)
        {
            this.Red = red;
            this.Green = green;
            this.Blue = blue;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="RGB"/> class.
        /// </summary>
        /// 
        /// <param name="color">Initialize from specified <see cref="System.Drawing.Color">color.</see></param>
        /// 
        public RGB(System.Drawing.Color color)
        {
            this.Red = color.R;
            this.Green = color.G;
            this.Blue = color.B;
        }
    }

    /// <summary>
    /// HSL components.
    /// </summary>
    /// 
    /// <remarks>The class encapsulates <b>HSL</b> color components.</remarks>
    /// 
    [Serializable]
    public class HSL
    {
        public static ImageFiltering filtering = new ImageFiltering();

        /// <summary>
        /// Hue component.
        /// </summary>
        /// 
        /// <remarks>Hue is measured in the range of [0, 359].</remarks>
        /// 
        public int Hue;

        /// <summary>
        /// Saturation component.
        /// </summary>
        /// 
        /// <remarks>Saturation is measured in the range of [0, 1].</remarks>
        /// 
        public double Saturation;

        /// <summary>
        /// Luminance value.
        /// </summary>
        /// 
        /// <remarks>Luminance is measured in the range of [0, 1].</remarks>
        /// 
        public double Luminance;

        public static void InitializeHSL(Settings settings)
        {
            try
            {
                HSL_RGB = new Dictionary<long, RGB>();
                settings.PropertyChanged += new System.ComponentModel.PropertyChangedEventHandler(Values_PropertyChanged);
                UpdateFromSettings(settings);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        public static bool _useHSLRounding = false;
        public static int _decimalDigitsHSLLuminanceRounding = 2;
        public static int _decimalDigitsHSLSaturationRounding = 2;

        public static void UpdateFromSettings(Settings settings)
        {
            _useHSLRounding = settings.UseHSLRounding;
            _decimalDigitsHSLLuminanceRounding = settings.DecimalDigitsHSLLuminanceRounding;
            _decimalDigitsHSLSaturationRounding = settings.DecimalDigitsHSLSaturationRounding;
        }

        static void Values_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            var settings = sender as Settings;
            UpdateFromSettings(settings);
            Interlocked.Exchange<Dictionary<long, RGB>>(ref HSL_RGB, new Dictionary<long, RGB>());
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="HSL"/> class.
        /// </summary>
        public HSL() { }

        /// <summary>
        /// Initializes a new instance of the <see cref="HSL"/> class.
        /// </summary>
        /// 
        /// <param name="hue">Hue component.</param>
        /// <param name="saturation">Saturation component.</param>
        /// <param name="luminance">Luminance component.</param>
        /// 
        public HSL(int hue, double saturation, double luminance)
        {
            this.Hue = hue;
            this.Saturation = saturation;
            this.Luminance = luminance;
        }

        public static Dictionary<int, HSL> RGB_HSL = new Dictionary<int, HSL>();

        /// <summary>
        /// Convert from RGB to HSL color space.
        /// </summary>
        /// 
        /// <param name="rgb">Source color in <b>RGB</b> color space.</param>
        /// <param name="hsl">Destination color in <b>HSL</b> color space.</param>
        /// 
        public unsafe static void FromRGB(RGB rgb, HSL hsl)
        {
            rgb.ToHSL(hsl);
            //int hue;
            //double sat;
            //double lum;
            //ImageFiltering.RGBToHSL(Red, Green, Blue, &hue, &sat, &lum);

            //hsl.Hue = hue;
            //hsl.Saturation = sat;
            //hsl.Luminance = lum;
            
            //ImageFiltering.RGBToHSL(rgb.Red, rgb.Green, rgb.Blue, &hsl.Hue, &hsl.Saturation, &hsl.Luminance);
            //HSL result;
            //int key = ((int)rgb.Red << 16) + ((int)rgb.Green << 8) + (int)rgb.Blue;

            //if (!RGB_HSL.TryGetValue(key, out result))
            //{
            //    result = new HSL();
            //    double r = (rgb.Red / 255.0);
            //    double g = (rgb.Green / 255.0);
            //    double b = (rgb.Blue / 255.0);

            //    double minRG;
            //    double maxRG;

            //    if (r > g)
            //    {
            //        minRG = g;
            //        maxRG = r;
            //    }
            //    else
            //    {
            //        minRG = r;
            //        maxRG = g;
            //    }

            //    double min = minRG > b ? b : minRG;// Math.Min(Math.Min(r, g), b);
            //    double max = maxRG > b ? maxRG : b;// Math.Max(Math.Max(r, g), b);
            //    double delta = max - min;

            //    // get luminance value
            //    result.Luminance = (max + min) / 2;

            //    if (delta == 0)
            //    {
            //        // gray color
            //        result.Hue = 0;
            //        result.Saturation = 0.0;
            //    }
            //    else
            //    {
            //        // get saturation value
            //        result.Saturation = (result.Luminance < 0.5) ? (delta / (max + min)) : (delta / (2 - max - min));

            //        // get hue value
            //        double del_r = (((max - r) / 6) + (delta / 2)) / delta;
            //        double del_g = (((max - g) / 6) + (delta / 2)) / delta;
            //        double del_b = (((max - b) / 6) + (delta / 2)) / delta;
            //        double hue;

            //        if (r == max)
            //            hue = del_b - del_g;
            //        else if (g == max)
            //            hue = (1.0 / 3) + del_r - del_b;
            //        else
            //            hue = (2.0 / 3) + del_g - del_r;

            //        // correct hue if needed
            //        if (hue < 0)
            //            hue += 1;
            //        if (hue > 1)
            //            hue -= 1;

            //        result.Hue = (int)(hue * 360);
            //    }

            //    try
            //    {
            //        if (result != null)
            //        {
            //            RGB_HSL.Add(key, result);
            //        }
            //    }
            //    catch
            //    {
            //        //Console.WriteLine("threading conflict with RGB to HSL");
            //    }
            //}

            //hsl.Hue = result.Hue;
            //hsl.Luminance = result.Luminance;
            //hsl.Saturation = result.Saturation;            
        }

        //public static HSL CalcAndAddHSLFromRGB(RGB rgb, int key)
        //{
        //    return CalcAndAddHSLFromRGB(rgb, key, true);
        //}

        //public static HSL CalcAndAddHSLFromRGB(RGB rgb, int key, bool add)
        //{
        //    HSL result = new HSL();
        //    double r = (rgb.Red / 255.0);
        //    double g = (rgb.Green / 255.0);
        //    double b = (rgb.Blue / 255.0);

        //    double minRG;
        //    double maxRG;

        //    if (r > g)
        //    {
        //        minRG = g;
        //        maxRG = r;
        //    }
        //    else
        //    {
        //        minRG = r;
        //        maxRG = g;
        //    }            

        //    double min = minRG > b ? b : minRG;// Math.Min(Math.Min(r, g), b);
        //    double max = maxRG > b ? maxRG : b;// Math.Max(Math.Max(r, g), b);
        //    double delta = max - min;

        //    // get luminance value
        //    result.Luminance = (max + min) / 2;

        //    if (delta == 0)
        //    {
        //        // gray color
        //        result.Hue = 0;
        //        result.Saturation = 0.0;
        //    }
        //    else
        //    {
        //        // get saturation value
        //        result.Saturation = (result.Luminance < 0.5) ? (delta / (max + min)) : (delta / (2 - max - min));

        //        // get hue value
        //        double del_r = (((max - r) / 6) + (delta / 2)) / delta;
        //        double del_g = (((max - g) / 6) + (delta / 2)) / delta;
        //        double del_b = (((max - b) / 6) + (delta / 2)) / delta;
        //        double hue;

        //        if (r == max)
        //            hue = del_b - del_g;
        //        else if (g == max)
        //            hue = (1.0 / 3) + del_r - del_b;
        //        else
        //            hue = (2.0 / 3) + del_g - del_r;

        //        // correct hue if needed
        //        if (hue < 0)
        //            hue += 1;
        //        if (hue > 1)
        //            hue -= 1;

        //        result.Hue = (int)(hue * 360);
        //    }

        //    try
        //    {
        //        if (result != null && add)
        //        {
        //            RGB_HSL.Add(key, result);
        //        }
        //    }
        //    catch
        //    {
        //        //Console.WriteLine("threading conflict with RGB to HSL");
        //    }

        //    return result;
        //}

        /// <summary>
        /// Convert from RGB to HSL color space.
        /// </summary>
        /// 
        /// <param name="rgb">Source color in <b>RGB</b> color space.</param>
        /// 
        /// <returns>Returns <see cref="HSL"/> instance, which represents converted color value.</returns>
        /// 
        public static HSL FromRGB(RGB rgb)
        {
            HSL hsl = new HSL();
            FromRGB(rgb, hsl);
            return hsl;
        }

        public static Dictionary<long, RGB> HSL_RGB = new Dictionary<long, RGB>();
        //public static Dictionary<double, RGB> HSL_RGB;

        /// <summary>
        /// Convert from HSL to RGB color space.
        /// </summary>
        /// 
        /// <param name="hsl">Source color in <b>HSL</b> color space.</param>
        /// <param name="rgb">Destination color in <b>RGB</b> color space.</param>
        /// 
        public unsafe static void ToRGB(HSL hsl, RGB rgb)
        {
            RGB temp;
            double key;

            if (_useHSLRounding)
            {
                // Make this a setting - 2 decimal point precision
                double lu = Math.Round(hsl.Luminance, _decimalDigitsHSLLuminanceRounding, MidpointRounding.AwayFromZero);
                double sat = Math.Round(hsl.Saturation, _decimalDigitsHSLSaturationRounding, MidpointRounding.AwayFromZero);

                key = (double)(hsl.Hue * 16) + (lu) + ((sat * 4D) + 1.1D);
            }
            else
            {
                //           leaves space of 16 between     [0,1]       [~2,~8]                
                // key = (double)(hsl.Hue * 16) + (hsl.Luminance) + ((hsl.Saturation * 4D) + 1.1D);
                key = (double)(hsl.Hue * 16) + (hsl.Luminance) + ((hsl.Saturation * 4D) + 1.1D);
            }

            long lKey;
            long* pl = (long*)((double*)&key);
            lKey = *pl;

            if (!HSL_RGB.TryGetValue(lKey, out temp))
            {
                temp = CalcRGBFromHSL(hsl);

                try
                {
                    if (temp != null)
                    {
                        HSL_RGB.Add(lKey, temp);
                    }
                }
                catch
                {
                    //Console.WriteLine("threading conflict with HSL to RGB");
                }
            }

            rgb.Red = temp.Red;
            rgb.Green = temp.Green;
            rgb.Blue = temp.Blue;
        }

        private static RGB CalcRGBFromHSL(HSL hsl)
        {
            RGB temp;

            if (hsl.Saturation == 0)
            {
                // gray values
                byte b = (byte)(hsl.Luminance * 255);
                temp = new RGB(b, b, b);
            }
            else
            {
                double v1, v2;
                double hue = (double)hsl.Hue / 360;

                v2 = (hsl.Luminance < 0.5) ?
                    (hsl.Luminance * (1 + hsl.Saturation)) :
                    ((hsl.Luminance + hsl.Saturation) - (hsl.Luminance * hsl.Saturation));
                v1 = 2 * hsl.Luminance - v2;

                temp = new RGB((byte)(255 * Hue_2_RGB(v1, v2, hue + (1.0 / 3))),
                               (byte)(255 * Hue_2_RGB(v1, v2, hue)),
                               (byte)(255 * Hue_2_RGB(v1, v2, hue - (1.0 / 3))));
            }

            return temp;
        }

        /// <summary>
        /// Convert the color to <b>RGB</b> color space.
        /// </summary>
        /// 
        /// <returns>Returns <see cref="RGB"/> instance, which represents converted color value.</returns>
        /// 
        public RGB ToRGB()
        {
            RGB rgb = new RGB();
            ToRGB(this, rgb);
            return rgb;
        }

        #region Private members

        private const double TwoThirds = 2.0D / 3.0D;

        // HSL to RGB helper routine
        private static double Hue_2_RGB(double v1, double v2, double vH)
        {
            if (vH < 0)
                vH += 1;
            else if (vH > 1)
                vH -= 1;

            if ((6 * vH) < 1)
                return (v1 + (v2 - v1) * 6 * vH);
            if ((2 * vH) < 1)
                return v2;
            if ((3 * vH) < 2)
                //return (v1 + (v2 - v1) * ((TwoThirds) - vH) * 6);
                return (v1 + (v2 - v1) * ((2.0 / 3) - vH) * 6);

            return v1;
        }
        #endregion
    }

    /// <summary>
    /// YCbCr components.
    /// </summary>
    /// 
    /// <remarks>The class encapsulates <b>YCbCr</b> color components.</remarks>
    /// 
    [Serializable]
    public class YCbCr
    {
        /// <summary>
        /// Index of <b>Y</b> component.
        /// </summary>
        public const short YIndex = 0;

        /// <summary>
        /// Index of <b>Cb</b> component.
        /// </summary>
        public const short CbIndex = 1;

        /// <summary>
        /// Index of <b>Cr</b> component.
        /// </summary>
        public const short CrIndex = 2;

        /// <summary>
        /// <b>Y</b> component.
        /// </summary>
        public double Y;

        /// <summary>
        /// <b>Cb</b> component.
        /// </summary>
        public double Cb;

        /// <summary>
        /// <b>Cr</b> component.
        /// </summary>
        public double Cr;

        /// <summary>
        /// Initializes a new instance of the <see cref="YCbCr"/> class.
        /// </summary>
        public YCbCr() { }

        /// <summary>
        /// Initializes a new instance of the <see cref="YCbCr"/> class.
        /// </summary>
        /// 
        /// <param name="y"><b>Y</b> component.</param>
        /// <param name="cb"><b>Cb</b> component.</param>
        /// <param name="cr"><b>Cr</b> component.</param>
        /// 
        public YCbCr(double y, double cb, double cr)
        {
            this.Y = Math.Max(0.0, Math.Min(1.0, y));
            this.Cb = Math.Max(-0.5, Math.Min(0.5, cb));
            this.Cr = Math.Max(-0.5, Math.Min(0.5, cr));
        }

        /// <summary>
        /// Convert from RGB to YCbCr color space (Rec 601-1 specification). 
        /// </summary>
        /// 
        /// <param name="rgb">Source color in <b>RGB</b> color space.</param>
        /// <param name="ycbcr">Destination color in <b>YCbCr</b> color space.</param>
        /// 
        public static void FromRGB(RGB rgb, YCbCr ycbcr)
        {
            double r = (double)rgb.Red / 255;
            double g = (double)rgb.Green / 255;
            double b = (double)rgb.Blue / 255;

            ycbcr.Y = 0.2989 * r + 0.5866 * g + 0.1145 * b;
            ycbcr.Cb = -0.1687 * r - 0.3313 * g + 0.5000 * b;
            ycbcr.Cr = 0.5000 * r - 0.4184 * g - 0.0816 * b;
        }

        /// <summary>
        /// Convert from RGB to YCbCr color space (Rec 601-1 specification).
        /// </summary>
        /// 
        /// <param name="rgb">Source color in <b>RGB</b> color space.</param>
        /// 
        /// <returns>Returns <see cref="YCbCr"/> instance, which represents converted color value.</returns>
        /// 
        public static YCbCr FromRGB(RGB rgb)
        {
            YCbCr ycbcr = new YCbCr();
            FromRGB(rgb, ycbcr);
            return ycbcr;
        }

        /// <summary>
        /// Convert from YCbCr to RGB color space.
        /// </summary>
        /// 
        /// <param name="ycbcr">Source color in <b>YCbCr</b> color space.</param>
        /// <param name="rgb">Destination color in <b>RGB</b> color spacs.</param>
        /// 
        public static void ToRGB(YCbCr ycbcr, RGB rgb)
        {
            // don't warry about zeros. compiler will remove them
            double r = Math.Max(0.0, Math.Min(1.0, ycbcr.Y + 0.0000 * ycbcr.Cb + 1.4022 * ycbcr.Cr));
            double g = Math.Max(0.0, Math.Min(1.0, ycbcr.Y - 0.3456 * ycbcr.Cb - 0.7145 * ycbcr.Cr));
            double b = Math.Max(0.0, Math.Min(1.0, ycbcr.Y + 1.7710 * ycbcr.Cb + 0.0000 * ycbcr.Cr));

            rgb.Red = (byte)(r * 255);
            rgb.Green = (byte)(g * 255);
            rgb.Blue = (byte)(b * 255);
        }

        /// <summary>
        /// Convert the color to <b>RGB</b> color space.
        /// </summary>
        /// 
        /// <returns>Returns <see cref="RGB"/> instance, which represents converted color value.</returns>
        /// 
        public RGB ToRGB()
        {
            RGB rgb = new RGB();
            ToRGB(this, rgb);
            return rgb;
        }
    }
}
