// AForge Image Processing Library
// AForge.NET framework
//
// Copyright  Andrew Kirillov, 2005-2009
// andrew.kirillov@aforgenet.com
//

namespace AForge.Imaging.Filters
{
    using System;
    using System.Collections.Generic;
    using System.Drawing;
    using System.Drawing.Imaging;
    using AForge;

    /// <summary>
    /// Linear correction of RGB channels.
    /// </summary>
    /// 
    /// <remarks><para>The filter performs linear correction of RGB channels by mapping specified
    /// channels' input ranges to output ranges. It is similar to the
    /// <see cref="ColorRemapping"/>, but the remapping is linear.</para>
    /// 
    /// <para>The filter accepts 8 bpp grayscale and 24/32 bpp color images for processing.</para>
    /// 
    /// <para>Sample usage:</para>
    /// <code>
    /// // create filter
    /// LevelsLinear filter = new LevelsLinear( );
    /// // set ranges
    /// filter.InRed   = new IntRange( 30, 230 );
    /// filter.InGreen = new IntRange( 50, 240 );
    /// filter.InBlue  = new IntRange( 10, 210 );
    /// // apply the filter
    /// filter.ApplyInPlace( Image );
    /// </code>
    /// 
    /// <para><b>Initial image:</b></para>
    /// <img src="img/imaging/sample1.jpg" width="480" height="361" />
    /// <para><b>Result image:</b></para>
    /// <img src="img/imaging/levels_linear.jpg" width="480" height="361" />
    /// </remarks>
    /// 
    /// <seealso cref="HSLLinear"/>
    /// <seealso cref="YCbCrLinear"/>
    /// 
    [Serializable] public class LevelsLinear : BaseInPlacePartialFilter
    {
        private IntRange inRed = new IntRange(0, 255);
        private IntRange inGreen = new IntRange(0, 255);
        private IntRange inBlue = new IntRange(0, 255);

        private IntRange outRed = new IntRange(0, 255);
        private IntRange outGreen = new IntRange(0, 255);
        private IntRange outBlue = new IntRange(0, 255);

        private byte[] mapRed = new byte[256];
        private byte[] mapGreen = new byte[256];
        private byte[] mapBlue = new byte[256];

        // private format translation dictionary
        private Dictionary<PixelFormat, PixelFormat> formatTransalations = new Dictionary<PixelFormat, PixelFormat>();

        /// <summary>
        /// Format translations dictionary.
        /// </summary>
        public override Dictionary<PixelFormat, PixelFormat> FormatTransalations
        {
            get { return formatTransalations; }
        }

        #region Public Propertis

        /// <summary>
        /// Red component's input range.
        /// </summary>
        public IntRange InRed
        {
            get { return inRed; }
            set
            {
                if (inRed != value)
                {
                    inRed = value;
                    CalculateMap(inRed, outRed, mapRed);
                    OnPropertyChanged("InRed");
                    OnPropertyChanged("InRedMin");
                    OnPropertyChanged("InRedMax");
                    OnPropertyChanged("InRedLength");
                }
            }
        }

        public int InRedMin
        {
            get { return InRed.Min; }
            set
            {
                if (InRed.Min != value)
                {
                    InRed = new IntRange(value, InRed.Max);
                }
            }
        }

        public int InRedMax
        {
            get { return InRed.Max; }
            set
            {
                if (InRed.Max != value)
                {
                    InRed = new IntRange(InRed.Min, value);
                }
            }
        }

        public int InRedLength
        {
            get { return InRed.Length; }
        }

        /// <summary>
        /// Green component's input range.
        /// </summary>
        public IntRange InGreen
        {
            get { return inGreen; }
            set
            {
                if (inGreen != value)
                {
                    inGreen = value;
                    CalculateMap(inGreen, outGreen, mapGreen);
                    OnPropertyChanged("InGreen");
                    OnPropertyChanged("InGreenMin");
                    OnPropertyChanged("InGreenMax");
                    OnPropertyChanged("InGreenLength");
                }
            }
        }


        public int InGreenMin
        {
            get { return InGreen.Min; }
            set
            {
                if (InGreen.Min != value)
                {
                    InGreen = new IntRange(value, InGreen.Max);
                }
            }
        }

        public int InGreenMax
        {
            get { return InGreen.Max; }
            set
            {
                if (InGreen.Max != value)
                {
                    InGreen = new IntRange(InGreen.Min, value);
                }
            }
        }

        public int InGreenLength
        {
            get { return InGreen.Length; }
        }

        /// <summary>
        /// Blue component's input range.
        /// </summary>
        public IntRange InBlue
        {
            get { return inBlue; }
            set
            {
                if (inBlue != value)
                {
                    inBlue = value;
                    CalculateMap(inBlue, outBlue, mapBlue);
                    OnPropertyChanged("InBlue");
                    OnPropertyChanged("InBlueMin");
                    OnPropertyChanged("InBlueMax");
                    OnPropertyChanged("InBlueLength");
                }
            }
        }

        public int InBlueMin
        {
            get { return InBlue.Min; }
            set
            {
                if (InBlue.Min != value)
                {
                    InBlue = new IntRange(value, InBlue.Max);
                }
            }
        }

        public int InBlueMax
        {
            get { return InBlue.Max; }
            set
            {
                if (InBlue.Max != value)
                {
                    InBlue = new IntRange(InBlue.Min, value);
                }
            }
        }

        public int InBlueLength
        {
            get { return InBlue.Length; }
        }
        /// <summary>
        /// Gray component's input range.
        /// </summary>
        public IntRange InGray
        {
            get { return inGreen; }
            set
            {
                if (inGreen != value)
                {
                    inGreen = value;
                    CalculateMap(inGreen, outGreen, mapGreen);
                    OnPropertyChanged("InGray");
                    OnPropertyChanged("InGrayMin");
                    OnPropertyChanged("InGrayMax");
                    OnPropertyChanged("InGrayLength");
                }
            }
        }

        public int InGrayMin
        {
            get { return InGray.Min; }
            set
            {
                if (InGray.Min != value)
                {
                    InGray = new IntRange(value, InGray.Max);
                }
            }
        }

        public int InGrayMax
        {
            get { return InGray.Max; }
            set
            {
                if (InGray.Max != value)
                {
                    InGray = new IntRange(InGray.Min, value);
                }
            }
        }

        public int InGrayLength
        {
            get { return InGray.Length; }
        }

        /// <summary>
        /// Input range for RGB components.
        /// </summary>
        /// 
        /// <remarks>The property allows to set red, green and blue input ranges to the same value.</remarks>
        /// 
        public IntRange Input
        {
            set
            {
                inRed = inGreen = inBlue = value;
                CalculateMap(inRed, outRed, mapRed);
                CalculateMap(inGreen, outGreen, mapGreen);
                CalculateMap(inBlue, outBlue, mapBlue);
            }
        }

        public IntRange OutRed
        {
            get { return outRed; }
            set
            {
                if (outRed != value)
                {
                    outRed = value;
                    CalculateMap(outRed, outRed, mapRed);
                    OnPropertyChanged("OutRed");
                    OnPropertyChanged("OutRedMin");
                    OnPropertyChanged("OutRedMax");
                    OnPropertyChanged("OutRedLength");
                }
            }
        }

        public int OutRedMin
        {
            get { return OutRed.Min; }
            set
            {
                if (OutRed.Min != value)
                {
                    OutRed = new IntRange(value, OutRed.Max);
                }
            }
        }

        public int OutRedMax
        {
            get { return OutRed.Max; }
            set
            {
                if (OutRed.Max != value)
                {
                    OutRed = new IntRange(OutRed.Min, value);
                }
            }
        }

        public int OutRedLength
        {
            get { return OutRed.Length; }
        }

        /// <summary>
        /// Green component's output range.
        /// </summary>
        public IntRange OutGreen
        {
            get { return outGreen; }
            set
            {
                if (outGreen != value)
                {
                    outGreen = value;
                    CalculateMap(outGreen, outGreen, mapGreen);
                    OnPropertyChanged("OutGreen");
                    OnPropertyChanged("OutGreenMin");
                    OnPropertyChanged("OutGreenMax");
                    OnPropertyChanged("OutGreenLength");
                }
            }
        }


        public int OutGreenMin
        {
            get { return OutGreen.Min; }
            set
            {
                if (OutGreen.Min != value)
                {
                    OutGreen = new IntRange(value, OutGreen.Max);
                }
            }
        }

        public int OutGreenMax
        {
            get { return OutGreen.Max; }
            set
            {
                if (OutGreen.Max != value)
                {
                    OutGreen = new IntRange(OutGreen.Min, value);
                }
            }
        }

        public int OutGreenLength
        {
            get { return OutGreen.Length; }
        }

        /// <summary>
        /// Blue component's output range.
        /// </summary>
        public IntRange OutBlue
        {
            get { return outBlue; }
            set
            {
                if (outBlue != value)
                {
                    outBlue = value;
                    CalculateMap(outBlue, outBlue, mapBlue);
                    OnPropertyChanged("OutBlue");
                    OnPropertyChanged("OutBlueMin");
                    OnPropertyChanged("OutBlueMax");
                    OnPropertyChanged("OutBlueLength");
                }
            }
        }

        public int OutBlueMin
        {
            get { return OutBlue.Min; }
            set
            {
                if (OutBlue.Min != value)
                {
                    OutBlue = new IntRange(value, OutBlue.Max);
                }
            }
        }

        public int OutBlueMax
        {
            get { return OutBlue.Max; }
            set
            {
                if (OutBlue.Max != value)
                {
                    OutBlue = new IntRange(OutBlue.Min, value);
                }
            }
        }

        public int OutBlueLength
        {
            get { return OutBlue.Length; }
        }
        /// <summary>
        /// Gray component's output range.
        /// </summary>
        public IntRange OutGray
        {
            get { return outGreen; }
            set
            {
                if (outGreen != value)
                {
                    outGreen = value;
                    CalculateMap(inGreen, outGreen, mapGreen);
                    OnPropertyChanged("OutGray");
                    OnPropertyChanged("OutGrayMin");
                    OnPropertyChanged("OutGrayMax");
                    OnPropertyChanged("OutGrayLength");
                }
            }
        }

        public int OutGrayMin
        {
            get { return OutGray.Min; }
            set
            {
                if (OutGray.Min != value)
                {
                    OutGray = new IntRange(value, OutGray.Max);
                }
            }
        }

        public int OutGrayMax
        {
            get { return OutGray.Max; }
            set
            {
                if (OutGray.Max != value)
                {
                    OutGray = new IntRange(OutGray.Min, value);
                }
            }
        }

        public int OutGrayLength
        {
            get { return OutGray.Length; }
        }
        /// <summary>
        /// Output range for RGB components.
        /// </summary>
        /// 
        /// <remarks>The property allows to set red, green and blue output ranges to the same value.</remarks>
        /// 
        public IntRange Output
        {
            set
            {
                OutRed = OutGreen = OutBlue = value;
                CalculateMap(inRed, outRed, mapRed);
                CalculateMap(inGreen, outGreen, mapGreen);
                CalculateMap(inBlue, outBlue, mapBlue);
            }
        }

        #endregion


        /// <summary>
        /// Initializes a new instance of the <see cref="LevelsLinear"/> class.
        /// </summary>
        public LevelsLinear()
        {
            CalculateMap(inRed, outRed, mapRed);
            CalculateMap(inGreen, outGreen, mapGreen);
            CalculateMap(inBlue, outBlue, mapBlue);

            formatTransalations[PixelFormat.Format8bppIndexed] = PixelFormat.Format8bppIndexed;
            formatTransalations[PixelFormat.Format24bppRgb] = PixelFormat.Format24bppRgb;
            formatTransalations[PixelFormat.Format32bppRgb] = PixelFormat.Format32bppRgb;
            formatTransalations[PixelFormat.Format32bppArgb] = PixelFormat.Format32bppArgb;
        }

        /// <summary>
        /// Process the filter on the specified image.
        /// </summary>
        /// 
        /// <param name="image">Source image data.</param>
        /// <param name="rect">Image rectangle for processing by the filter.</param>
        ///
        protected override unsafe void ProcessFilter(UnmanagedImage image, Rectangle rect)
        {
            int pixelSize = Image.GetPixelFormatSize(image.PixelFormat) / 8;

            // processing start and stop X,Y positions
            int startX = rect.Left;
            int startY = rect.Top;
            int stopX = startX + rect.Width;
            int stopY = startY + rect.Height;
            int offset = image.Stride - rect.Width * pixelSize;

            // do the job
            byte* ptr = (byte*)image.ImageData.ToPointer();

            // allign pointer to the first pixel to process
            ptr += (startY * image.Stride + startX * pixelSize);

            if (image.PixelFormat == PixelFormat.Format8bppIndexed)
            {
                // grayscale image
                for (int y = startY; y < stopY; y++)
                {
                    for (int x = startX; x < stopX; x++, ptr++)
                    {
                        // gray
                        *ptr = mapGreen[*ptr];
                    }
                    ptr += offset;
                }
            }
            else
            {
                // RGB image
                for (int y = startY; y < stopY; y++)
                {
                    for (int x = startX; x < stopX; x++, ptr += pixelSize)
                    {
                        // red
                        ptr[RGB.R] = mapRed[ptr[RGB.R]];
                        // green
                        ptr[RGB.G] = mapGreen[ptr[RGB.G]];
                        // blue
                        ptr[RGB.B] = mapBlue[ptr[RGB.B]];
                    }
                    ptr += offset;
                }
            }
        }


        /// <summary>
        /// Calculate conversion map.
        /// </summary>
        /// 
        /// <param name="inRange">Input range.</param>
        /// <param name="outRange">Output range.</param>
        /// <param name="map">Conversion map.</param>
        /// 
        private void CalculateMap(IntRange inRange, IntRange outRange, byte[] map)
        {
            double k = 0, b = 0;

            if (inRange.Max != inRange.Min)
            {
                k = (double)(outRange.Max - outRange.Min) / (double)(inRange.Max - inRange.Min);
                b = (double)(outRange.Min) - k * inRange.Min;
            }

            for (int i = 0; i < 256; i++)
            {
                byte v = (byte)i;

                if (v >= inRange.Max)
                    v = (byte)outRange.Max;
                else if (v <= inRange.Min)
                    v = (byte)outRange.Min;
                else
                    v = (byte)(k * v + b);

                map[i] = v;
            }
        }
    }
}
