// AForge Image Processing Library
// AForge.NET framework
//
// Copyright  Andrew Kirillov, 2005-2008
// andrew.kirillov@aforgenet.com
//

namespace AForge.Imaging.Filters
{
    using System;
    using System.Collections.Generic;
    using System.Drawing;
    using System.Drawing.Imaging;
    using CPPClassLibrary;

    /// <summary>
    /// Hue modifier.
    /// </summary>
    /// 
    /// <remarks><para>The filter operates in <b>HSL</b> color space and updates
    /// pixels' hue values setting it to the specified value (luminance and
    /// saturation are kept unchanged). The result of the filter looks like the image
    /// is observed through a glass of the given color.</para>
    ///
    /// <para>The filter accepts 24 and 32 bpp color images for processing.</para>
    /// <para>Sample usage:</para>
    /// <code>
    /// // create filter
    /// HueModifier filter = new HueModifier( 180 );
    /// // apply the filter
    /// filter.ApplyInPlace( image );
    /// </code>
    /// 
    /// <para><b>Initial image:</b></para>
    /// <img src="img/imaging/sample1.jpg" width="480" height="361" />
    /// <para><b>Result image:</b></para>
    /// <img src="img/imaging/hue_modifier.jpg" width="480" height="361" />
    /// </remarks>
    /// 
    [Serializable]
    public class HueModifier : BaseInPlacePartialFilter
    {
        public static ImageFiltering filtering = new ImageFiltering();
        private int hue = 0;

        // private format translation dictionary
        private Dictionary<PixelFormat, PixelFormat> formatTransalations = new Dictionary<PixelFormat, PixelFormat>();

        /// <summary>
        /// Format translations dictionary.
        /// </summary>
        public override Dictionary<PixelFormat, PixelFormat> FormatTransalations
        {
            get { return formatTransalations; }
        }

        /// <summary>
        /// Hue value to set, [0, 359].
        /// </summary>
        /// 
        /// <remarks><para>Default value is set to <b>0</b>.</para></remarks>
        /// 
        public int Hue
        {
            get { return hue; }
            set
            {
                int temp = Math.Max(0, Math.Min(359, value));

                if (temp != hue)
                {
                    hue = temp;
                    OnPropertyChanged("Hue");
                }
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="HueModifier"/> class.
        /// </summary>
        /// 
        public HueModifier()
        {
            formatTransalations[PixelFormat.Format24bppRgb] = PixelFormat.Format24bppRgb;
            formatTransalations[PixelFormat.Format32bppRgb] = PixelFormat.Format32bppRgb;
            formatTransalations[PixelFormat.Format32bppArgb] = PixelFormat.Format32bppArgb;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="HueModifier"/> class.
        /// </summary>
        /// 
        /// <param name="hue">Hue value to set.</param>
        /// 
        public HueModifier(int hue)
            : this()
        {
            this.hue = hue;
        }

        /// <summary>
        /// Process the filter on the specified image.
        /// </summary>
        /// 
        /// <param name="image">Source image data.</param>
        /// <param name="rect">Image rectangle for processing by the filter.</param>
        ///
        protected override unsafe void ProcessFilter(UnmanagedImage image, Rectangle rect)
        {
            int startX = rect.Left;
            int startY = rect.Top;
            int stopX = startX + rect.Width;
            int stopY = startY + rect.Height;
            int offset = image.Stride - rect.Width * 3;

            RGB rgb = new RGB();
            HSL hsl = new HSL();

            // do the job
            byte* ptr = (byte*)image.ImageData.ToPointer();

            // allign pointer to the first pixel to process
            ptr += (startY * image.Stride + startX * 3);

            // for each row
            for (int y = startY; y < stopY; y++)
            {
                // for each pixel
                for (int x = startX; x < stopX; x++, ptr += 3)
                {
                    rgb.Red = ptr[RGB.R];
                    rgb.Green = ptr[RGB.G];
                    rgb.Blue = ptr[RGB.B];

                    //int key = ((int)rgb.Red << 16) + ((int)rgb.Green << 8) + (int)rgb.Blue;
                    //HSL temp;

                    int hueOut;
                    double sat;
                    double lum;
                    ImageFiltering.DoRGBToHSL(rgb.Red, rgb.Green, rgb.Blue, &hueOut, &sat, &lum);

                    //// convert to HSL                    
                    //if (!AForge.Imaging.HSL.RGB_HSL.TryGetValue(key, out temp))
                    //{
                    //    temp = AForge.Imaging.HSL.CalcAndAddHSLFromRGB(rgb, key);
                    //}

                    //if (temp == null)
                    //{
                    //    temp = AForge.Imaging.HSL.CalcAndAddHSLFromRGB(rgb, key, false);
                    //}

                    //hsl.Luminance = temp.Luminance;
                    //hsl.Saturation = temp.Saturation;

                    //// modify hue value
                    //hsl.Hue = hue;

                    hsl.Luminance = lum;
                    hsl.Saturation = sat;

                    // modify hue value
                    hsl.Hue = hue;

                    // convert back to RGB
                    AForge.Imaging.HSL.ToRGB(hsl, rgb);

                    ptr[RGB.R] = rgb.Red;
                    ptr[RGB.G] = rgb.Green;
                    ptr[RGB.B] = rgb.Blue;
                }
                ptr += offset;
            }
        }
    }
}
