﻿// Original author contact info: Owen Emlen (owene_1998@yahoo.com)
// Note: other individuals may also have contributed to this code
// Project hosted on CodePlex.com as of 1/10/2009 at http://www.codeplex.com/EmlenMud
using System;
using System.Linq;
using System.Xml.Serialization;
using System.Collections;
using System.Runtime.Serialization;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Threading;
#if USE_HYPER
using Hyper.ComponentModel;
#endif

namespace BrainTechLLC.ThreadSafeObjects
{
    /// <summary>
    /// A multilevel dictionary provides quick access to items given a sequence of (multiple) keys
    /// </summary>    
#if NO_SILVERLIGHT
    [TypeDescriptionProvider(typeof(HyperTypeDescriptionProvider))]
#endif
    [Serializable]
    [DataContract]
    public class MultiLevelDictionary<TKey, TItem> : INotifyCollectionChanged
    {
        [DataMember]
        public Dictionary<TKey, MultiLevelDictionary<TKey, TItem>> ItemLookup;

        [DataMember]
        public Dictionary<TKey, TItem> TerminatingItems;

        [NonSerialized, XmlIgnore]
        public Cached<List<KeyValuePair<List<TKey>, TItem>>> AllTerminatingItemsListCache;

        /// <summary>
        /// Returns a count of all items in the multi-level dictionary
        /// </summary>
        public int ItemCount
        {
            get
            {
                int count = 0;

                count += (TerminatingItems == null) ? 0 : TerminatingItems.Count;

                if (ItemLookup != null)
                    ItemLookup.ForEach((dict) => count += dict.Value.ItemCount);

                return count;
            }
        }

        /// <summary>
        /// Returns the max depth (max length of sequence of keys) that define the longest key
        /// </summary>
        /// <param name="startDepth"></param>
        /// <returns></returns>
        public int MaxDepth(int startDepth)
        {
            int max = startDepth;
            if (ItemLookup != null)
            {
                foreach (var kvp in ItemLookup)
                {
                    if (kvp.Value != null)
                    {
                        int depth = kvp.Value.MaxDepth(startDepth + 1);
                        if (depth > max) max = depth;
                    }
                }
            }
            return max;
        }

        /// <summary>
        /// Executes an action on all items at a given level, or executes levelTooDeep if the current level has no item lookup
        /// </summary>
        /// <param name="level"></param>
        /// <param name="currentLevel"></param>
        /// <param name="action"></param>
        /// <param name="levelTooDeep"></param>
        public void AtLevel(int level, int currentLevel, Action<Dictionary<TKey, TItem>> action, Action<int> levelTooDeep)
        {
            if (currentLevel == level && TerminatingItems != null)
            {
                action(TerminatingItems);
            }
            else
            {
                if (ItemLookup == null)
                {
                    levelTooDeep(currentLevel + 1);
                }
                else
                {
                    foreach (var v in ItemLookup)
                    {
                        if (v.Value != null)
                            v.Value.AtLevel(level, currentLevel + 1, action, levelTooDeep);
                    }
                }
            }
        }

        /// <summary>
        /// Returns all items, regardless of key
        /// </summary>
        public Dictionary<List<TKey>, TItem> AllTerminatingItems
        {
            get { return ReadTerminatingItemsRecursive(new List<TKey>()); }
        }

        /// <summary>
        /// Returns all items at all depths/levels
        /// </summary>
        public List<TItem> AllItems
        {
            get
            {
                List<TItem> items = new List<TItem>();
                var all = AllTerminatingItemsList;

                for (int n = 0; n < all.Count; n++)
                {
                    if (all[n].Value != null)
                    {
                        items.Add(all[n].Value);
                    }
                }
                return items;
            }
        }

        /// <summary>
        /// Returns a cached key/value pair list of keys and corresponding items at all depths/levels
        /// </summary>
        public List<KeyValuePair<List<TKey>, TItem>> AllTerminatingItemsList
        {
            get
            {
                if (AllTerminatingItemsListCache == null)
                {
                    Interlocked.CompareExchange<Cached<List<KeyValuePair<List<TKey>, TItem>>>>(
                        ref AllTerminatingItemsListCache,
                         new Cached<List<KeyValuePair<List<TKey>, TItem>>>(
                             delegate(object parameterIn, out List<KeyValuePair<List<TKey>, TItem>> results, object[] args)
                             {
                                 List<KeyValuePair<List<TKey>, TItem>> valuePairs = new List<KeyValuePair<List<TKey>, TItem>>();
                                 Dictionary<List<TKey>, TItem> items = ReadTerminatingItemsRecursive(new List<TKey>());

                                 foreach (KeyValuePair<List<TKey>, TItem> item in items)
                                     valuePairs.Add(item);

                                 results = valuePairs;
                                 return true;
                             }), null);
                }
                return AllTerminatingItemsListCache.GetItem(null, null);
            }
        }

        /// <summary>
        /// Determines if the multi-level dictionary contains a specific item
        /// </summary>
        /// <param name="item"></param>
        /// <returns></returns>
        public bool ContainsValue(TItem item)
        {
            if (TerminatingItems != null)
            {
                bool contains = TerminatingItems.ContainsValue(item);

                if (contains)
                    return true;
            }

            if (ItemLookup != null)
            {
                foreach (var v in ItemLookup)
                {
                    bool contains = v.Value.ContainsValue(item);

                    if (contains)
                        return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Reads all terminating items
        /// To get key values, keeps track of all keys for each level visited (kinda expensive)
        /// </summary>
        /// <param name="keysSoFar"></param>
        /// <returns></returns>
        public Dictionary<List<TKey>, TItem> ReadTerminatingItemsRecursive(List<TKey> keysSoFar)
        {
            Dictionary<List<TKey>, TItem> allItems = new Dictionary<List<TKey>, TItem>();

            TerminatingItems.ForEachIfNotNull((item) =>
                 {
                     List<TKey> keysSoFarCopy = new List<TKey>(keysSoFar);
                     keysSoFarCopy.Add(item.Key);
                     allItems.Add(keysSoFarCopy, item.Value);
                 });

            if (ItemLookup != null)
            {
                foreach (KeyValuePair<TKey, MultiLevelDictionary<TKey, TItem>> item in ItemLookup)
                {
                    List<TKey> keysSoFarCopy = new List<TKey>(keysSoFar);
                    keysSoFarCopy.Add(item.Key);
                    Dictionary<List<TKey>, TItem> subItems = item.Value.ReadTerminatingItemsRecursive(keysSoFarCopy);
                    subItems.ForEachIfNotNull(subItem => allItems.Add(subItem.Key, subItem.Value));
                }
            }

            return allItems;
        }

        /// <summary>
        /// Removes a terminating item by key with a maximum key level specified
        /// </summary>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        /// <returns></returns>
        public bool RemoveTerminatingItem(int keyLevel, int maxLevel, params TKey[] keys)
        {
            TKey key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                TItem item = default(TItem);
                if (TerminatingItems != null && TerminatingItems.TryGetValue(key, out item))
                {
                    // possible race condition here
                    TerminatingItems.Remove(key);
                    AllTerminatingItemsListCache.SafeMarkDirty();
                    OnNotifyCollectionChanged(NotifyCollectionChangedAction.Remove, item);
                    return true;
                }

                return false;
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelDictionary<TKey, TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                    {
                        bool found = dictionary.RemoveTerminatingItem(keyLevel + 1, maxLevel, keys);

                        if (found)
                            AllTerminatingItemsListCache.SafeMarkDirty();

                        return found;
                    }
                }

                return false;
            }
        }

        /// <summary>
        /// Removes a terminating item by key with a maximum key level specified
        /// </summary>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        /// <returns></returns>
        public bool RemoveTerminatingItem(int keyLevel, int maxLevel, List<TKey> keys)
        {
            TKey key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                TItem item = default(TItem);
                if (TerminatingItems != null && TerminatingItems.TryGetValue(key, out item))
                {
                    // possible race condition here
                    TerminatingItems.Remove(key);
                    AllTerminatingItemsListCache.SafeMarkDirty();
                    OnNotifyCollectionChanged(NotifyCollectionChangedAction.Remove, item);
                    return true;
                }

                return false;
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelDictionary<TKey, TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                    {
                        bool found = dictionary.RemoveTerminatingItem(keyLevel + 1, maxLevel, keys);

                        if (found)
                            AllTerminatingItemsListCache.SafeMarkDirty();

                        return found;
                    }
                }

                return false;
            }
        }

        /// <summary>
        /// Removes a terminating item by key
        /// </summary>
        /// <param name="keys"></param>
        /// <returns></returns>
        public bool RemoveTerminatingItem(params TKey[] keys)
        {
            int keysLength = keys.Length;

            if (keysLength == 0)
            {
                keys = new TKey[] { default(TKey) };
                keysLength = 1;
                //throw new Exception("Must have at least 1 key");
            }

            return RemoveTerminatingItem(0, keysLength, keys);
        }

        /// <summary>
        /// Removes a terminating item by key
        /// </summary>
        /// <param name="keys"></param>
        /// <returns></returns>
        public bool RemoveTerminatingItem(List<TKey> keys)
        {
            int keysLength = keys.Count;

            if (keysLength == 0)
            {
                keys = new List<TKey>() { default(TKey) };
                keysLength = 1;
            }

            return RemoveTerminatingItem(0, keysLength, keys);
        }

        /// <summary>
        /// Adds a terminating item by key
        /// </summary>
        /// <param name="item"></param>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        public void AddTerminatingItem(TItem item, int keyLevel, int maxLevel, params TKey[] keys)
        {
            TKey key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                Populate.OnDemand(ref TerminatingItems);

                if (!TerminatingItems.ContainsKey(key))
                {
                    // TODO: Think through race condition where key added twice
                    try { TerminatingItems.Add(key, item); }
                    catch { AddTerminatingItem(item, keyLevel, maxLevel, keys); }
                    OnNotifyCollectionChanged(NotifyCollectionChangedAction.Add, item);
                }
                else
                {
                    TItem old = default(TItem);
                    if (TerminatingItems.TryGetValue(key, out old))
                    {
                        TerminatingItems[key] = item;

                        if (CollectionChanged != null)
                            OnNotifyCollectionChanged(NotifyCollectionChangedAction.Replace, new List<TItem>() { old, item });
                    }
                }
            }
            else
            {
                Populate.OnDemand(ref ItemLookup);

                MultiLevelDictionary<TKey, TItem> dictionary;

                if (!ItemLookup.TryGetValue(key, out dictionary))
                {
                    MultiLevelDictionary<TKey, TItem> newDictionary = new MultiLevelDictionary<TKey, TItem>();

                    // TODO: Think through race condition where key added twice
                    try
                    {
                        ItemLookup.Add(key, newDictionary);
                        newDictionary.AddTerminatingItem(item, keyLevel + 1, maxLevel, keys);
                    }
                    catch { AddTerminatingItem(item, keyLevel, maxLevel, keys); }
                }
                else { dictionary.AddTerminatingItem(item, keyLevel + 1, maxLevel, keys); }
            }

            AllTerminatingItemsListCache.SafeMarkDirty();
        }

        /// <summary>
        /// Adds a terminating item by key
        /// </summary>
        /// <param name="item"></param>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        public void AddTerminatingItem(TItem item, int keyLevel, int maxLevel, List<TKey> keys)
        {
            TKey key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                Populate.OnDemand(ref TerminatingItems);

                if (!TerminatingItems.ContainsKey(key))
                {
                    // TODO: Think through race condition where key added twice
                    try { TerminatingItems.Add(key, item); }
                    catch { AddTerminatingItem(item, keyLevel, maxLevel, keys); }
                    OnNotifyCollectionChanged(NotifyCollectionChangedAction.Add, item);
                }
                else
                {
                    TItem old = default(TItem);
                    if (TerminatingItems.TryGetValue(key, out old))
                    {
                        TerminatingItems[key] = item;

                        if (CollectionChanged != null)
                            OnNotifyCollectionChanged(NotifyCollectionChangedAction.Replace, new List<TItem>() { old, item });
                    }
                }
            }
            else
            {
                Populate.OnDemand(ref ItemLookup);

                MultiLevelDictionary<TKey, TItem> dictionary;

                if (!ItemLookup.TryGetValue(key, out dictionary))
                {
                    MultiLevelDictionary<TKey, TItem> newDictionary = new MultiLevelDictionary<TKey, TItem>();

                    // TODO: Think through race condition where key added twice
                    try
                    {
                        ItemLookup.Add(key, newDictionary);
                        newDictionary.AddTerminatingItem(item, keyLevel + 1, maxLevel, keys);
                    }
                    catch { AddTerminatingItem(item, keyLevel, maxLevel, keys); }
                }
                else { dictionary.AddTerminatingItem(item, keyLevel + 1, maxLevel, keys); }
            }

            AllTerminatingItemsListCache.SafeMarkDirty();
        }

        /// <summary>
        /// Adds a terminating item by key
        /// </summary>
        /// <param name="item"></param>
        /// <param name="keys"></param>
        public void AddTerminatingItem(TItem item, params TKey[] keys)
        {
            int keysLength = keys.Length;

            if (keysLength == 0)
            {
                keys = new TKey[] { default(TKey) };
                keysLength = 1;
                //throw new Exception("Must have at least 1 key");
            }

            AddTerminatingItem(item, 0, keysLength, keys);
        }

        /// <summary>
        /// Adds a terminating item by key
        /// </summary>
        /// <param name="item"></param>
        /// <param name="keys"></param>
        public void AddTerminatingItem(TItem item, List<TKey> keys)
        {
            int keysLength = keys.Count;

            if (keysLength == 0)
            {
                keys = new List<TKey>() { default(TKey) };
                keysLength = 1;
            }

            AddTerminatingItem(item, 0, keysLength, keys);
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        /// <returns></returns>
        public TItem FindItem(int keyLevel, int maxLevel, params TKey[] keys)
        {
            TKey key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                if (TerminatingItems != null)
                {
                    TItem item;
                    if (TerminatingItems.TryGetValue(key, out item))
                        return item;
                }

                return default(TItem);
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelDictionary<TKey, TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                        return dictionary.FindItem(keyLevel + 1, maxLevel, keys);
                }

                return default(TItem);
            }
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        /// <returns></returns>
        public TItem FindItem(int keyLevel, int maxLevel, List<TKey> keys)
        {
            TKey key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                if (TerminatingItems != null)
                {
                    TItem item;
                    if (TerminatingItems.TryGetValue(key, out item))
                        return item;
                }

                return default(TItem);
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelDictionary<TKey, TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                        return dictionary.FindItem(keyLevel + 1, maxLevel, keys);
                }

                return default(TItem);
            }
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="keys"></param>
        /// <returns></returns>
        public TItem FindItem(params TKey[] keys)
        {
            int keysLength = keys.Length;

            if (keysLength == 0)
            {
                keys = new TKey[] { default(TKey) };
                keysLength = 1;
            }

            return FindItem(0, keysLength, keys);
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="keys"></param>
        /// <returns></returns>
        public TItem FindItem(List<TKey> keys)
        {
            int keysLength = keys.Count;

            if (keysLength == 0)
            {
                keys = new List<TKey>() { default(TKey) };
                keysLength = 1;
            }

            return FindItem(0, keysLength, keys);
        }

        public Dictionary<TKey, TItem> FindItemsWithKeyPrefix(int keyLevel, params TKey[] keys)
        {
            // UNTESTED
            if (keys == null || keys.Length == 0)
                return null;

            TKey key = keys[keyLevel];
            int maxLevel = keys.Length;

            if (keyLevel == maxLevel)
            {
                if (TerminatingItems != null)
                {
                    // create a new one
                    return TerminatingItems;
                }

                return null;
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelDictionary<TKey, TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                        return FindItemsWithKeyPrefix(keyLevel + 1, keys);
                }

                return null;
            }
        }

        public void OnNotifyCollectionChanged(NotifyCollectionChangedAction action, IList<TItem> changedItems)
        {
#if SILVERLIGHT
#else
            if (CollectionChanged != null)
            {
                switch (action)
                {
                    case NotifyCollectionChangedAction.Add:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TItem>(action, changedItems));
                        break;
                    case NotifyCollectionChangedAction.Remove:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TItem>(action, changedItems));
                        break;
                    case NotifyCollectionChangedAction.Replace:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TItem>(action, changedItems[0], changedItems[1]));
                        break;
                    case NotifyCollectionChangedAction.Reset:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TItem>(action));
                        break;
                }
            }
#endif
        }

        public void OnNotifyCollectionChanged(NotifyCollectionChangedAction action, TItem changedItem)
        {
            if (CollectionChanged != null)
                OnNotifyCollectionChanged(action, new List<TItem>() { changedItem });
        }

        #region INotifyCollectionChanged Members

        [field: NonSerialized]
        public event NotifyCollectionChangedEventHandler CollectionChanged;

        #endregion
    }

    /// <summary>
    /// Special case Multi-level dictionary that is indexed by a sequence of integers
    /// </summary>
    /// <typeparam name="TItem"></typeparam>
    [Serializable]
    [DataContract]
    public class MultiLevelCustomDictionary<TItem> : INotifyCollectionChanged
    {
        [DataMember]
        public Dictionary<int, MultiLevelCustomDictionary<TItem>> ItemLookup;

        [DataMember]
        public Dictionary<int, TItem> TerminatingItems;

        [NonSerialized, XmlIgnore]
        public Cached<List<KeyValuePair<List<int>, TItem>>> AllTerminatingItemsListCache;

        /// <summary>
        /// Returns a count of all items in the multi-level dictionary
        /// </summary>
        public int ItemCount
        {
            get
            {
                int count = 0;

                count += (TerminatingItems == null) ? 0 : TerminatingItems.Count;

                if (ItemLookup != null)
                    ItemLookup.ForEach((dict) => count += dict.Value.ItemCount);

                return count;
            }
        }

        /// <summary>
        /// Returns the max depth (max length of sequence of integers) that define the longest key
        /// </summary>
        /// <param name="startDepth"></param>
        /// <returns></returns>
        public int MaxDepth(int startDepth)
        {
            int max = startDepth;
            if (ItemLookup != null)
            {
                foreach (var kvp in ItemLookup)
                {
                    if (kvp.Value != null)
                    {
                        int depth = kvp.Value.MaxDepth(startDepth + 1);
                        if (depth > max) max = depth;
                    }
                }
            }
            return max;
        }

        /// <summary>
        /// Executes an action on all items at a given level, or executes levelTooDeep if the current level has no item lookup
        /// </summary>
        /// <param name="level"></param>
        /// <param name="currentLevel"></param>
        /// <param name="action"></param>
        /// <param name="levelTooDeep"></param>
        public void AtLevel(int level, int currentLevel, Action<Dictionary<int, TItem>> action, Action<int> levelTooDeep)
        {
            if (currentLevel == level && TerminatingItems != null)
            {
                action(TerminatingItems);
            }
            else
            {
                if (ItemLookup == null)
                {
                    levelTooDeep(currentLevel + 1);
                }
                else
                {
                    foreach (var v in ItemLookup)
                    {
                        if (v.Value != null)
                        {
                            v.Value.AtLevel(level, currentLevel + 1, action, levelTooDeep);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Returns all items, regardless of key
        /// </summary>
        public Dictionary<List<int>, TItem> AllTerminatingItems
        {
            get { return ReadTerminatingItemsRecursive(new List<int>()); }
        }

        /// <summary>
        /// Returns all items at all depths/levels
        /// </summary>
        public List<TItem> AllItems
        {
            get
            {
                List<TItem> items = new List<TItem>();
                var all = AllTerminatingItemsList;

                for (int n = 0; n < all.Count; n++)
                {
                    if (all[n].Value != null)
                    {
                        items.Add(all[n].Value);
                    }
                }
                return items;
            }
        }

        /// <summary>
        /// Returns a cached key/value pair list of keys and corresponding items at all depths/levels
        /// </summary>
        public List<KeyValuePair<List<int>, TItem>> AllTerminatingItemsList
        {
            get
            {
                if (AllTerminatingItemsListCache == null)
                {
                    Interlocked.CompareExchange<Cached<List<KeyValuePair<List<int>, TItem>>>>(
                        ref AllTerminatingItemsListCache,
                         new Cached<List<KeyValuePair<List<int>, TItem>>>(
                             delegate(object parameterIn, out List<KeyValuePair<List<int>, TItem>> results, object[] args)
                             {
                                 List<KeyValuePair<List<int>, TItem>> valuePairs = new List<KeyValuePair<List<int>, TItem>>();
                                 Dictionary<List<int>, TItem> items = ReadTerminatingItemsRecursive(new List<int>());

                                 foreach (KeyValuePair<List<int>, TItem> item in items)
                                     valuePairs.Add(item);

                                 results = valuePairs;
                                 return true;
                             }), null);
                }
                return AllTerminatingItemsListCache.GetItem(null, null);
            }
        }

        /// <summary>
        /// Determines if the multi-level dictionary contains a specific item
        /// </summary>
        /// <param name="item"></param>
        /// <returns></returns>
        public bool ContainsValue(TItem item)
        {
            if (TerminatingItems != null)
            {
                bool contains = TerminatingItems.ContainsValue(item);

                if (contains)
                    return true;
            }

            if (ItemLookup != null)
            {
                foreach (var v in ItemLookup)
                {
                    bool contains = v.Value.ContainsValue(item);

                    if (contains)
                        return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Reads all terminating items
        /// To get key values, keeps track of all keys for each level visited (kinda expensive)
        /// </summary>
        /// <param name="keysSoFar"></param>
        /// <returns></returns>
        public Dictionary<List<int>, TItem> ReadTerminatingItemsRecursive(List<int> keysSoFar)
        {
            Dictionary<List<int>, TItem> allItems = new Dictionary<List<int>, TItem>();

            TerminatingItems.ForEachIfNotNull((item) =>
            {
                List<int> keysSoFarCopy = new List<int>(keysSoFar);
                keysSoFarCopy.Add(item.Key);
                allItems.Add(keysSoFarCopy, item.Value);
            });

            if (ItemLookup != null)
            {
                foreach (KeyValuePair<int, MultiLevelCustomDictionary<TItem>> item in ItemLookup)
                {
                    List<int> keysSoFarCopy = new List<int>(keysSoFar);
                    keysSoFarCopy.Add(item.Key);
                    Dictionary<List<int>, TItem> subItems = item.Value.ReadTerminatingItemsRecursive(keysSoFarCopy);
                    subItems.ForEachIfNotNull(subItem => allItems.Add(subItem.Key, subItem.Value));
                }
            }

            return allItems;
        }

        /// <summary>
        /// Removes a terminating item by key with a maximum key level specified
        /// </summary>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        /// <returns></returns>
        public bool RemoveTerminatingItem(int keyLevel, int maxLevel, params int[] keys)
        {
            int key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                TItem item = default(TItem);
                if (TerminatingItems != null && TerminatingItems.TryGetValue(key, out item))
                {
                    // possible race condition here
                    TerminatingItems.Remove(key);
                    if (AllTerminatingItemsListCache != null)
                        AllTerminatingItemsListCache.SafeMarkDirty();
                    OnNotifyCollectionChanged(NotifyCollectionChangedAction.Remove, item);
                    return true;
                }

                return false;
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelCustomDictionary<TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                    {
                        bool found = dictionary.RemoveTerminatingItem(keyLevel + 1, maxLevel, keys);

                        if (found)
                            AllTerminatingItemsListCache.SafeMarkDirty();

                        return found;
                    }
                }

                return false;
            }
        }

        /// <summary>
        /// Removes a terminating item by key with a maximum key level specified
        /// </summary>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        /// <returns></returns>
        public bool RemoveTerminatingItem(int keyLevel, int maxLevel, List<int> keys)
        {
            int key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                TItem item = default(TItem);
                if (TerminatingItems != null && TerminatingItems.TryGetValue(key, out item))
                {
                    // possible race condition here
                    TerminatingItems.Remove(key);
                    AllTerminatingItemsListCache.SafeMarkDirty();
                    OnNotifyCollectionChanged(NotifyCollectionChangedAction.Remove, item);
                    return true;
                }

                return false;
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelCustomDictionary<TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                    {
                        bool found = dictionary.RemoveTerminatingItem(keyLevel + 1, maxLevel, keys);

                        if (found)
                            AllTerminatingItemsListCache.SafeMarkDirty();

                        return found;
                    }
                }

                return false;
            }
        }

        /// <summary>
        /// Removes a terminating item by key
        /// </summary>
        /// <param name="keys"></param>
        /// <returns></returns>
        public bool RemoveTerminatingItem(params int[] keys)
        {
            int keysLength = keys.Length;

            if (keysLength == 0)
            {
                keys = new int[] { default(int) };
                keysLength = 1;
                //throw new Exception("Must have at least 1 key");
            }

            return RemoveTerminatingItem(0, keysLength, keys);
        }

        /// <summary>
        /// Removes a terminating item by key
        /// </summary>
        /// <param name="keys"></param>
        /// <returns></returns>
        public bool RemoveTerminatingItem(List<int> keys)
        {
            int keysLength = keys.Count;

            if (keysLength == 0)
            {
                keys = new List<int>() { default(int) };
                keysLength = 1;
            }

            return RemoveTerminatingItem(0, keysLength, keys);
        }

        /// <summary>
        /// Adds a terminating item by key
        /// </summary>
        /// <param name="item"></param>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        public void AddTerminatingItem(TItem item, int keyLevel, int maxLevel, params int[] keys)
        {
            int key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                Populate.OnDemand(ref TerminatingItems);

                if (!TerminatingItems.ContainsKey(key))
                {
                    // TODO: Think through race condition where key added twice
                    try { TerminatingItems.Add(key, item); }
                    catch { AddTerminatingItem(item, keyLevel, maxLevel, keys); }
                    OnNotifyCollectionChanged(NotifyCollectionChangedAction.Add, item);
                }
                else
                {
                    TItem old = default(TItem);
                    if (TerminatingItems.TryGetValue(key, out old))
                    {
                        TerminatingItems[key] = item;

                        if (CollectionChanged != null)
                            OnNotifyCollectionChanged(NotifyCollectionChangedAction.Replace, new List<TItem>() { old, item });
                    }
                }
            }
            else
            {
                Populate.OnDemand(ref ItemLookup);

                MultiLevelCustomDictionary<TItem> dictionary;

                if (!ItemLookup.TryGetValue(key, out dictionary))
                {
                    MultiLevelCustomDictionary<TItem> newDictionary = new MultiLevelCustomDictionary<TItem>();

                    // TODO: Think through race condition where key added twice
                    try
                    {
                        ItemLookup.Add(key, newDictionary);
                        newDictionary.AddTerminatingItem(item, keyLevel + 1, maxLevel, keys);
                    }
                    catch { AddTerminatingItem(item, keyLevel, maxLevel, keys); }
                }
                else { dictionary.AddTerminatingItem(item, keyLevel + 1, maxLevel, keys); }
            }

            AllTerminatingItemsListCache.SafeMarkDirty();
        }

        /// <summary>
        /// Adds a terminating item by key
        /// </summary>
        /// <param name="item"></param>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        public void AddTerminatingItem(TItem item, int keyLevel, int maxLevel, List<int> keys)
        {
            int key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                Populate.OnDemand(ref TerminatingItems);

                if (!TerminatingItems.ContainsKey(key))
                {
                    // TODO: Think through race condition where key added twice
                    try { TerminatingItems.Add(key, item); }
                    catch { AddTerminatingItem(item, keyLevel, maxLevel, keys); }
                    OnNotifyCollectionChanged(NotifyCollectionChangedAction.Add, item);
                }
                else
                {
                    TItem old = default(TItem);
                    if (TerminatingItems.TryGetValue(key, out old))
                    {
                        TerminatingItems[key] = item;

                        if (CollectionChanged != null)
                            OnNotifyCollectionChanged(NotifyCollectionChangedAction.Replace, new List<TItem>() { old, item });
                    }
                }
            }
            else
            {
                Populate.OnDemand(ref ItemLookup);

                MultiLevelCustomDictionary<TItem> dictionary;

                if (!ItemLookup.TryGetValue(key, out dictionary))
                {
                    MultiLevelCustomDictionary<TItem> newDictionary = new MultiLevelCustomDictionary<TItem>();

                    // TODO: Think through race condition where key added twice
                    try
                    {
                        ItemLookup.Add(key, newDictionary);
                        newDictionary.AddTerminatingItem(item, keyLevel + 1, maxLevel, keys);
                    }
                    catch { AddTerminatingItem(item, keyLevel, maxLevel, keys); }
                }
                else { dictionary.AddTerminatingItem(item, keyLevel + 1, maxLevel, keys); }
            }

            AllTerminatingItemsListCache.SafeMarkDirty();
        }

        /// <summary>
        /// Adds a terminating item by keys x, y, z
        /// </summary>
        /// <param name="item"></param>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="z"></param>
        public void AddTerminatingItem(TItem item, int x, int y, int z)
        {
            AddTerminatingItem(item, 0, 3, x, y, z);
        }

        /// <summary>
        /// Adds a terminating item by key
        /// </summary>
        /// <param name="item"></param>
        /// <param name="keys"></param>
        public void AddTerminatingItem(TItem item, params int[] keys)
        {
            int keysLength = keys.Length;

            if (keysLength == 0)
            {
                keys = new int[] { default(int) };
                keysLength = 1;
                //throw new Exception("Must have at least 1 key");
            }

            AddTerminatingItem(item, 0, keysLength, keys);
        }

        /// <summary>
        /// Adds a terminating item by key
        /// </summary>
        /// <param name="item"></param>
        /// <param name="keys"></param>
        public void AddTerminatingItem(TItem item, List<int> keys)
        {
            int keysLength = keys.Count;

            if (keysLength == 0)
            {
                keys = new List<int>() { default(int) };
                keysLength = 1;
            }

            AddTerminatingItem(item, 0, keysLength, keys);
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        /// <returns></returns>
        public TItem FindItem(int keyLevel, int maxLevel, params int[] keys)
        {
            int key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                if (TerminatingItems != null)
                {
                    TItem item;
                    if (TerminatingItems.TryGetValue(key, out item))
                        return item;
                }

                return default(TItem);
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelCustomDictionary<TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                        return dictionary.FindItem(keyLevel + 1, maxLevel, keys);
                }

                return default(TItem);
            }
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="keyLevel"></param>
        /// <param name="maxLevel"></param>
        /// <param name="keys"></param>
        /// <returns></returns>
        public TItem FindItem(int keyLevel, int maxLevel, List<int> keys)
        {
            int key = keys[keyLevel];

            if (keyLevel == maxLevel - 1)
            {
                if (TerminatingItems != null)
                {
                    TItem item;
                    if (TerminatingItems.TryGetValue(key, out item))
                        return item;
                }

                return default(TItem);
            }
            else
            {
                if (ItemLookup != null)
                {
                    MultiLevelCustomDictionary<TItem> dictionary;

                    if (ItemLookup.TryGetValue(key, out dictionary))
                        return dictionary.FindItem(keyLevel + 1, maxLevel, keys);
                }

                return default(TItem);
            }
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="keys"></param>
        /// <returns></returns>
        public TItem FindItem(params int[] keys)
        {
            int keysLength = keys.Length;

            if (keysLength == 0)
            {
                keys = new int[] { default(int) };
                keysLength = 1;
            }

            return FindItem(0, keysLength, keys);
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="z"></param>
        /// <returns></returns>
        public TItem FindItem(int x, int y, int z)
        {
            return FindItem(0, 3, x, y, z);
        }

        /// <summary>
        /// Finds a terminating item by key
        /// </summary>
        /// <param name="keys"></param>
        /// <returns></returns>
        public TItem FindItem(List<int> keys)
        {
            int keysLength = keys.Count;

            if (keysLength == 0)
            {
                keys = new List<int>() { default(int) };
                keysLength = 1;
            }

            return FindItem(0, keysLength, keys);
        }

        public void OnNotifyCollectionChanged(NotifyCollectionChangedAction action, IList<TItem> changedItems)
        {
#if SILVERLIGHT
#else
            if (CollectionChanged != null)
            {
                switch (action)
                {
                    case NotifyCollectionChangedAction.Add:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TItem>(action, changedItems));
                        break;
                    case NotifyCollectionChangedAction.Remove:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TItem>(action, changedItems));
                        break;
                    case NotifyCollectionChangedAction.Replace:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TItem>(action, changedItems[0], changedItems[1]));
                        break;
                    case NotifyCollectionChangedAction.Reset:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TItem>(action));
                        break;
                }
            }
#endif
        }

        public void OnNotifyCollectionChanged(NotifyCollectionChangedAction action, TItem changedItem)
        {
            if (CollectionChanged != null)
                OnNotifyCollectionChanged(action, new List<TItem>() { changedItem });
        }

        #region INotifyCollectionChanged Members

        [field: NonSerialized]
        public event NotifyCollectionChangedEventHandler CollectionChanged;

        #endregion
    }

#if NO_SILVERLIGHT
    [TypeDescriptionProvider(typeof(HyperTypeDescriptionProvider))]
#endif
    [Serializable]
    [DataContract]
    public class MultiLevelDictionaryEntry<TKey, TItem>
    {
        [DataMember]
        public List<TKey> AllKeys;

        [DataMember]
        public TItem Item;

        public MultiLevelDictionaryEntry() { AllKeys = new List<TKey>(); }
    }

    /// <summary>
    /// A sparse array can be indexed by any number of indices (of type T1)
    /// Items in the array are of type T2
    /// </summary>
#if NO_SILVERLIGHT
    [TypeDescriptionProvider(typeof(HyperTypeDescriptionProvider))]
#endif
    [Serializable]
    [DataContract]
    public class SparseArray<T1, T2> : MultiLevelDictionary<T1, T2> where T2 : class
    {
        public SparseArray() { }

        public bool IsFixedSize { get { return false; } }
        public bool IsReadOnly { get { return false; } }
        public bool IsSynchronized { get { return false; } }
        public int Count { get { return ItemCount; } }
        public int Rank { get { return MaxDepth(0); } }
        public object SyncRoot { get { return null; } }

        public void CopyTo(Array array, int index)
        {
            throw new NotImplementedException();
        }

        public T1 GetLowerBound(int dimension)
        {
            return (from item in AllTerminatingItemsList where item.Key.Count > dimension select item.Key[dimension]).Min();
        }

        public T1 GetUpperBound(int dimension)
        {
            return (from item in AllTerminatingItemsList where item.Key.Count > dimension select item.Key[dimension]).Max();
        }

        public T2 GetValue(params T1[] indices) { return base.FindItem(indices); }
        public T2 GetValue(T1 index) { return GetValue(new T1[] { index }); }
        public T2 GetValue(T1 index1, T1 index2) { return GetValue(new T1[] { index1, index2 }); }

        public T2 GetValueOrCreate(Func<T2> create, params T1[] indices)
        {
            T2 item = GetValue(indices);

            if (item == null)
            {
                // RACE CONDITION?
                item = create();
                SetValue(item, indices);
            }

            return item;
        }

        public void SetValue(T2 value, params T1[] indices)
        {
            // RACE CONDITION?			
            if (value != null) { base.AddTerminatingItem((T2)value, indices); }
        }

        public void SetValue(T2 value, T1 index) { SetValue(value, new T1[] { index }); }
        public void SetValue(T2 value, T1 index1, T1 index2) { SetValue(value, new T1[] { index1, index2 }); }

        private class SparseArrayEnumerator : IEnumerator<T2>
        {
            IEnumerator _enum;
            SparseArray<T1, T2> _parent;
            List<KeyValuePair<List<T1>, T2>> _list;

            public void Dispose() { }

            public SparseArrayEnumerator(SparseArray<T1, T2> array)
            {
                _parent = array;
                _list = (from item in array.AllTerminatingItems select item).ToList();
                _enum = _list.GetEnumerator();
            }

            public void Reset() { _enum.Reset(); }
            public bool MoveNext() { return _enum.MoveNext(); }
            public object Current { get { return _enum.Current; } }
            T2 System.Collections.Generic.IEnumerator<T2>.Current { get { return (T2)_enum.Current; } }
            public int[] Index { get { return ((KeyValuePair<List<int>, T2>)Current).Key.ToArray(); } }
        }

        public System.Collections.IEnumerator GetEnumerator() { return new SparseArrayEnumerator(this); }
        public void RemoveAt(int index) { throw new NotImplementedException(); }
        public void Insert(int index, object value) { throw new NotImplementedException(); }
        public void Remove(T2 value) { }
        public bool Contains(object value) { return base.ContainsValue((T2)value); }

        public void Clear()
        {
            if (base.TerminatingItems != null)
            {
                base.TerminatingItems.Clear();
            }

            if (base.ItemLookup != null)
            {
                base.ItemLookup.Clear();
            }
        }

        public int IndexOf(T2 value)
        {
            return 0;
        }

        public int Add(T2 value)
        {
            throw new NotImplementedException();
        }

        public T2 this[params T1[] indicies]
        {
            get { return GetValue(indicies); }
            set { SetValue(value, indicies); }
        }

        public T2 this[T1 index]
        {
            get { return GetValue(index); }
            set { SetValue(value, index); }
        }

        public T2 this[T1 index1, T1 index2]
        {
            get { return GetValue(index1, index2); }
            set { SetValue(value, index1, index2); }
        }
    }

    /// <summary>
    /// Special case sparse array where array indices are integers
    /// </summary>
    [Serializable]
    [DataContract]
    public class SparseArrayCustom<T2> : MultiLevelCustomDictionary<T2> where T2 : class
    {
        public SparseArrayCustom() { }

        public bool IsFixedSize { get { return false; } }
        public bool IsReadOnly { get { return false; } }
        public bool IsSynchronized { get { return false; } }
        public int Count { get { return ItemCount; } }
        public int Rank { get { return MaxDepth(0); } }
        public object SyncRoot { get { return null; } }

        public void CopyTo(Array array, int index)
        {
            throw new NotImplementedException();
        }

        public int GetLowerBound(int dimension)
        {
            return (from item in AllTerminatingItemsList where item.Key.Count > dimension select item.Key[dimension]).Min();
        }

        public int GetUpperBound(int dimension)
        {
            return (from item in AllTerminatingItemsList where item.Key.Count > dimension select item.Key[dimension]).Max();
        }

        public T2 GetValue(params int[] indices) { return base.FindItem(indices); }
        public T2 GetValue(int index) { return GetValue(new int[] { index }); }
        public T2 GetValue(int index1, int index2) { return GetValue(new int[] { index1, index2 }); }

        public T2 GetValueOrCreate(Func<T2> create, params int[] indices)
        {
            T2 item = GetValue(indices);

            if (item == null)
            {
                // RACE CONDITION?
                item = create();
                SetValue(item, indices);
            }

            return item;
        }

        public void SetValue(T2 value, params int[] indices)
        {
            // RACE CONDITION?			
            if (value != null) { base.AddTerminatingItem((T2)value, indices); }
        }

        public void SetValue(T2 value, int index) { SetValue(value, new int[] { index }); }
        public void SetValue(T2 value, int index1, int index2) { SetValue(value, new int[] { index1, index2 }); }

        private class SparseArrayEnumerator : IEnumerator<T2>
        {
            IEnumerator _enum;
            SparseArrayCustom<T2> _parent;
            List<KeyValuePair<List<int>, T2>> _list;

            public void Dispose() { }

            public SparseArrayEnumerator(SparseArrayCustom<T2> array)
            {
                _parent = array;
                _list = (from item in array.AllTerminatingItems select item).ToList();
                _enum = _list.GetEnumerator();
            }

            public void Reset() { _enum.Reset(); }
            public bool MoveNext() { return _enum.MoveNext(); }
            public object Current { get { return _enum.Current; } }
            T2 System.Collections.Generic.IEnumerator<T2>.Current { get { return (T2)_enum.Current; } }
            public int[] Index { get { return ((KeyValuePair<List<int>, T2>)Current).Key.ToArray(); } }
        }

        public System.Collections.IEnumerator GetEnumerator() { return new SparseArrayEnumerator(this); }
        public void RemoveAt(int index) { throw new NotImplementedException(); }
        public void Insert(int index, object value) { throw new NotImplementedException(); }
        public void Remove(T2 value) { }
        public bool Contains(object value) { return base.ContainsValue((T2)value); }

        public void Clear()
        {
            base.TerminatingItems.Clear();
            base.ItemLookup.Clear();
        }

        public int IndexOf(T2 value)
        {
            return 0;
        }

        public int Add(T2 value)
        {
            throw new NotImplementedException();
        }

        public T2 this[params int[] indicies]
        {
            get { return GetValue(indicies); }
            set { SetValue(value, indicies); }
        }

        public T2 this[int index]
        {
            get { return GetValue(index); }
            set { SetValue(value, index); }
        }

        public T2 this[int index1, int index2]
        {
            get { return GetValue(index1, index2); }
            set { SetValue(value, index1, index2); }
        }

    }
}