﻿// Original author contact info: Owen Emlen (owene_1998@yahoo.com)
// Note: other individuals may also have contributed to this code
// Project hosted on CodePlex.com as of 1/10/2009 at http://www.codeplex.com/EmlenMud
using System;
using System.Xml.Serialization;
using System.Runtime.Serialization;
using System.Linq;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Collections;
using System.ComponentModel;
#if USE_HYPER
using Hyper.ComponentModel;
#endif

namespace BrainTechLLC.ThreadSafeObjects
{
    /// <summary>
    /// Thread-safe lookup allows multiple threads to look up items by key, add items,
    /// and remove items in a thread-safe manner
    /// </summary>
    /// <typeparam name="TListType"></typeparam>
    /// <typeparam name="TIndex"></typeparam>
#if NO_SILVERLIGHT
    [TypeDescriptionProvider(typeof(HyperTypeDescriptionProvider))]
#endif
    [Serializable]
    [DataContract]
    public class ThreadSafeLookupNonRef<TIndex, TType> : Lockable, INotifyCollectionChanged,
        IDictionary<TIndex, TType>, IMultipleItems<TType>, ISupportsCount
    {
        public override string ToString()
        {
            return Count.ToString();
        }

        // add track changes
        [NonSerialized, XmlIgnore]
        public bool Modified;

        [DataMember]
        public Dictionary<TIndex, TType> Lookup = new Dictionary<TIndex, TType>();

        /// <summary>
        /// Returns an array containing all items in the collection
        /// </summary>
        [XmlIgnore]
        public TType[] ArrayOfItems { get { return AllItems.ToArray(); } }

        /// <summary>
        /// Returns a mostly-accurate count of items currently in the collection
        /// </summary>
        [XmlIgnore]
        public int Count { get { return Lookup.Count; } }

        public bool ContainsKey(TIndex key) { return Lookup.ContainsKey(key); }

        public bool TryGetValue(TIndex key, out TType items) { return Lookup.TryGetValue(key, out items); }

        public bool Remove(KeyValuePair<TIndex, TType> kvp)
        {
            return Remove(kvp.Key, kvp.Value);
        }

        /// <summary>
        /// Returns a snapshot/list copy of all items in the collection
        /// </summary>
        public List<TType> AllItems
        {
            get
            {
                List<TType> items = new List<TType>(Lookup.Count);

                AquireLock();
                {
                    foreach (KeyValuePair<TIndex, TType> kvp in Lookup)
                    {
                        items.Add(kvp.Value);
                    }
                }
                ReleaseLock();

                return items;
            }
        }

        /// <summary>
        /// Adds an item under multiple keys
        /// </summary>
        /// <param name="keys"></param>
        /// <param name="item"></param>
        /// <returns></returns>
        public bool AddMultipleKeys(IEnumerable<TIndex> keys, TType item)
        {
            bool added = false;

            AquireLock();
            {
                foreach (TIndex key in keys)
                {
                    if (!Lookup.ContainsKey(key))
                    {
                        Lookup.Add(key, item);
                        added = true;
                    }
                }
            }
            ReleaseLock();

            if (added)
                OnNotifyCollectionChanged(NotifyCollectionChangedAction.Add, item);

            return added;
        }

        /// <summary>
        /// Adds multiple items according to their corresponding item keys
        /// </summary>
        /// <param name="keys"></param>
        /// <param name="values"></param>
        /// <returns></returns>
        public bool AddMultiple(IEnumerable<TIndex> keys, IList<TType> values)
        {
            List<TType> addedItems = null;
            bool added = false;
            int n = 0;

            if (CollectionChanged != null)
                addedItems = new List<TType>();

            AquireLock();
            {
                foreach (TIndex key in keys)
                {
                    if (!Lookup.ContainsKey(key))
                    {
                        Lookup.Add(key, values[n]);
                        if (addedItems != null) addedItems.Add(values[n]);
                        Modified = true;
                        added = true;
                    }
                    n++;
                }
            }
            ReleaseLock();

            if (added)
                OnNotifyCollectionChanged(NotifyCollectionChangedAction.Add, addedItems);

            return added;
        }

        /// <summary>
        /// Adds an item to the collection
        /// </summary>
        /// <returns>True if the item was already in the collection, otherwise false</returns>        
        public bool Add(TIndex key, TType item)
        {
            bool found = true;

            AquireLock();
            {
                if (!Lookup.ContainsKey(key))
                {
                    Lookup.Add(key, item);
                    Modified = true;
                    found = false;
                }
            }
            ReleaseLock();

            if (!found)
                OnNotifyCollectionChanged(NotifyCollectionChangedAction.Add, item);

            return !found;
        }

        /// <summary>
        /// If the item doesn't exist, run func(), add its return value
        /// </summary>
        /// <param name="key"></param>
        /// <param name="func"></param>
        /// <returns></returns>
        public bool AddIfNew(TIndex key, Func<TType> func)
        {
            if (!Lookup.ContainsKey(key))
                return Add(key, func());

            return false;
        }

        /// <summary>
        /// Adds a value if it does not exist, otherwise sets the value having the specified key to 'item'
        /// </summary>
        /// <param name="key"></param>
        /// <param name="item"></param>
        public void AddOrSet(TIndex key, TType item)
        {
            if (Lookup.ContainsKey(key))
            {
                TType old = default(TType);
                Lookup.TryGetValue(key, out old);
                if (!old.Equals(item))
                {
                    Lookup[key] = item;
                    Modified = true;

                    if (CollectionChanged != null)
                        OnNotifyCollectionChanged(NotifyCollectionChangedAction.Replace, new List<TType>() { old, item });
                }
            }
            else
            {
                Add(key, item);
                OnNotifyCollectionChanged(NotifyCollectionChangedAction.Add, item);
            }
        }

        /// <summary>
        /// Gets the value with the corresponding key in the collection, or returns default(TListType)
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public TType this[TIndex key]
        {
            get
            {
                TType item;

                // Not totally safe for lookups that frequently add/remove items.
                // For thread safety, surround the following line with AquireLock(); / ReleaseLock();
                if (!Lookup.TryGetValue(key, out item)) item = default(TType);

                return item;
            }
            set
            {
                AddOrSet(key, value);
            }
        }

        /// <summary>
        /// Removes an item by key
        /// </summary>
        /// <param name="key"></param>
        /// <param name="item"></param>
        /// <returns></returns>
        public bool Remove(TIndex key, TType item)
        {
            return Remove(key);
        }

        /// <summary>
        /// Removes an item by key
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public bool Remove(TIndex key)
        {
            bool removed = false;

            if (!Lookup.ContainsKey(key))
                return false;

            TType item = default(TType);
            AquireLock();
            {
                if (Lookup.TryGetValue(key, out item))
                {
                    Lookup.Remove(key);
                    Modified = true;
                    removed = true;
                }
            }
            ReleaseLock();

            if (removed)
                OnNotifyCollectionChanged(NotifyCollectionChangedAction.Remove, item);

            return removed;
        }

        /// <summary>
        /// Clears all items from the collection
        /// </summary>
        public void Clear()
        {
            IList<TType> items = null;

            if (CollectionChanged != null)
                items = new List<TType>(AllItems);

            AquireLock();
            {
                Lookup.Clear();
                Modified = true;
            }
            ReleaseLock();

            OnNotifyCollectionChanged(NotifyCollectionChangedAction.Reset, items);
        }

        /// <summary>
        /// Merges two thread safe lookups
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="other"></param>
        /// <returns></returns>
        public ThreadSafeLookupNonRef<TIndex, TType> Merge<T>(ThreadSafeLookupNonRef<TIndex, T> other) where T : class, TType
        {
            var kvps = other.AllKeyValuePairs;
            AquireLock();
            foreach (var kvp in kvps)
            {
                if (!Lookup.ContainsKey(kvp.Key))
                {
                    Lookup.Add(kvp.Key, kvp.Value);
                }
            }
            ReleaseLock();

            return this;
        }

        /// <summary>
        /// Performs action on each item (locks before iteration, unlocks after)
        /// </summary>
        /// <param name="action"></param>
        public void ForEachLocked(Action<TIndex, TType> action)
        {
            AquireLock();
            {
                var kvps = AllKeyValuePairs;
                foreach (var kvp in kvps)
                {
                    action(kvp.Key, kvp.Value);
                }
            }
            ReleaseLock();
        }

        /// <summary>
        /// Returns all key/value pairs in the lookup
        /// </summary>
        [XmlIgnore]
        public List<KeyValuePair<TIndex, TType>> AllKeyValuePairs
        {
            get
            {
                List<KeyValuePair<TIndex, TType>> results = new List<KeyValuePair<TIndex, TType>>();
                AquireLock();
                {
                    foreach (var v in Lookup)
                        results.Add(v);
                }
                ReleaseLock();

                return results;
            }
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return Lookup.GetEnumerator();
        }

        public IEnumerator<KeyValuePair<TIndex, TType>> GetEnumerator()
        {
            return Lookup.GetEnumerator();
        }

        public bool IsReadOnly { get { return false; } }

        /// <summary>
        /// Returns all keys for items in the collection
        /// </summary>
        public List<TIndex> Keys
        {
            get
            {
                List<TIndex> items = new List<TIndex>(Lookup.Count);
                AquireLock(); { Lookup.ForEach(kvp => items.Add(kvp.Key)); } ReleaseLock();
                return items;
            }
        }

        public void OnNotifyCollectionChanged(NotifyCollectionChangedAction action, IList<TType> changedItems)
        {
#if SILVERLIGHT
#else
            if (CollectionChanged != null)
            {
                switch (action)
                {
                    case NotifyCollectionChangedAction.Add:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TType>(action, changedItems));
                        break;
                    case NotifyCollectionChangedAction.Remove:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TType>(action, changedItems));
                        break;
                    case NotifyCollectionChangedAction.Replace:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TType>(action, changedItems[0], changedItems[1]));
                        break;
                    case NotifyCollectionChangedAction.Reset:
                        CollectionChanged(this, new NotifyCollectionChangedEventArgsEx<TType>(action));
                        break;
                }
            }
#endif
        }

        public void OnNotifyCollectionChanged(NotifyCollectionChangedAction action, TType changedItem)
        {
            if (CollectionChanged != null)
                OnNotifyCollectionChanged(action, new List<TType>() { changedItem });
        }

        #region INotifyCollectionChanged Members

        [field: NonSerialized]
        public event NotifyCollectionChangedEventHandler CollectionChanged;

        #endregion

        #region IDictionary<TIndex,TListType> Members

        void IDictionary<TIndex, TType>.Add(TIndex key, TType value)
        {
            this.Add(key, value);
        }

        ICollection<TIndex> IDictionary<TIndex, TType>.Keys
        {
            get { return Lookup.Keys; }
        }

        public ICollection<TType> Values
        {
            get { return Lookup.Values; }
        }

        #endregion

        #region ICollection<KeyValuePair<TIndex,TListType>> Members

        /// <summary>
        /// Adds an item to the lookup
        /// </summary>
        /// <param name="item"></param>
        public void Add(KeyValuePair<TIndex, TType> item)
        {
            Add(item.Key, item.Value);
        }

        public bool Contains(KeyValuePair<TIndex, TType> item)
        {
            return Lookup.Contains(item);
        }

        #endregion

        #region ICollection<KeyValuePair<TIndex,TListType>> Members

        public void CopyTo(KeyValuePair<TIndex, TType>[] array, int arrayIndex)
        {
            AquireLock();
            {
                foreach (KeyValuePair<TIndex, TType> item in Lookup)
                {
                    array[arrayIndex++] = item;
                }
            }
            ReleaseLock();
        }

        #endregion
    }
}